## The random forest code using CSS functions

import numpy
import pandas
import random
import datetime
import psycopg2
import pandas as pd


def entropy(data):
    _, uniqueClassesCounts = numpy.unique(data[:, -1], return_counts = True)
    probabilities = uniqueClassesCounts / uniqueClassesCounts.sum()
    return sum(probabilities * -numpy.log2(probabilities))

def information_gain(left, right):
    p_left = len(left) / (len(left) + len(right))
    p_right = len(right) / (len(left) + len(right))
    return p_left * entropy(right) + p_right * entropy(right)

def split_set(df, test_size):
    if isinstance(test_size, float):
        test_size = round(test_size * len(df))
    indices = df.index.tolist()
    i = random.sample(population = indices, k = test_size)
    df_test = df.loc[i]
    df_train = df.drop(i)
    return df_train, df_test

def check_purity(data):
    if len(numpy.unique(data[:, -1])) == 1:
        return True
    else:
        return False

def classify_data(data):
    unique_class, unique_lasses_count = numpy.unique(data[:, -1], return_counts = True)
    return unique_class[unique_lasses_count.argmax()]

def all__potential_splits(data, random_n):
    splits = {}
    _, columns = data.shape
    i = list(range(columns - 1))
    if random_n != None  and len(random_n) <= len(i):
        i = random_n
    for col in i:
        values = data[:, col]
        unique_val = numpy.unique(values)
        if len(unique_val) == 1:
            splits[col] = unique_val
        else:
            splits[col] = []
            for j in range(len(unique_val)):
                if j != 0:
                    currentValue = unique_val[j]
                    previousValue = unique_val[j - 1]
                    splits[col].append((currentValue + previousValue) / 2)
    return splits

def split_data(data, splitting_col, splitting_val):
    splitColumnValues = data[:, splitting_col]
    return data[splitColumnValues <= splitting_val], data[splitColumnValues > splitting_val]

def splitting_point(data, potential_splits, random_splits = None):
    total_IG = 9999
    splitting_col = 0
    splitting_val = 0
    if random_splits == None:
        for cols in potential_splits:
            for vals in potential_splits[cols]:
                left, right = split_data(data, cols, vals)
                current_IG = information_gain(left, right)
                if current_IG <= total_IG:
                    total_IG = current_IG
                    splitting_col = cols
                    splitting_val = vals
    else:
        for i in range(random_splits):
            random_col = random.choice(list(potential_splits))
            random_val = random.choice(potential_splits[random_col])
            left, right = split_data(data, random_col, random_val)
            current_IG = information_gain(left, right)
            if current_IG <= total_IG:
                total_IG = current_IG
                splitting_col = random_col
                splitting_val = random_val
    return splitting_col, splitting_val

def decision_tree(dataFrame, currentDepth = 0, minSampleSize = 2, maxDepth = 1000, randomAttributes = None, randomSplits = None):
    if currentDepth == 0:
        global COLUMN_HEADERS
        COLUMN_HEADERS = dataFrame.columns
        data = dataFrame.values
        if randomAttributes != None and randomAttributes <= len(COLUMN_HEADERS) - 1:
            randomAttributes = random.sample(population = list(range(len(COLUMN_HEADERS) - 1)), k = randomAttributes)
        else:
            randomAttributes = None
    else:
        data = dataFrame
    if check_purity(data) or len(data) < minSampleSize or currentDepth == maxDepth:
        return classify_data(data)
    else:
        currentDepth += 1
        potentialSplits = all__potential_splits(data, randomAttributes)
        splitColumn, splitValue = splitting_point(data, potentialSplits, randomSplits)
        dataBelow, dataAbove = split_data(data, splitColumn, splitValue)
        if len(dataBelow) == 0 or len(dataAbove) == 0:
            return classify_data(data)
        else:
            question = str(COLUMN_HEADERS[splitColumn]) + " <= " + str(splitValue)
            decisionSubTree = {question: []}
            yesAnswer = decision_tree(dataBelow, currentDepth, minSampleSize, maxDepth, randomAttributes, randomSplits)
            noAnswer = decision_tree(dataAbove, currentDepth, minSampleSize, maxDepth, randomAttributes, randomSplits)
            if yesAnswer == noAnswer:
                decisionSubTree = yesAnswer
            else:
                decisionSubTree[question].append(yesAnswer)
                decisionSubTree[question].append(noAnswer)
            return decisionSubTree

def classifySample(sample, decisionTree):
    if not isinstance(decisionTree, dict):
        return decisionTree
    question = list(decisionTree.keys())[0]
    attribute, value = question.split(" <= ")
    if sample[attribute] <= float(value):
        answer = decisionTree[question][0]
    else:
        answer = decisionTree[question][1]
    return classifySample(sample, answer)

def decision_tree_predict(dataFrame, decisionTree):
    predictions = dataFrame.apply(classifySample, axis = 1, args = (decisionTree,))
    return predictions

def calculateAccuracy(predictedResults, category):
    resultCorrect = predictedResults == category
    return resultCorrect.mean()

def make_bootstrap(train_df, bootstrap_size):
    i = numpy.random.randint(low = 0, high = len(train_df), size = bootstrap_size)
    return train_df.iloc[i]

def random_forest(train_df, bootstrap_size, randomAttributes, randomSplits, forestSize = 20, treeMaxDepth = 1000):
    forest = []
    for i in range(forestSize):
        bootstrappedDataFrame = make_bootstrap(train_df, bootstrap_size)
        decisionTree = decision_tree(bootstrappedDataFrame, maxDepth = treeMaxDepth, randomAttributes = randomAttributes,
                                         randomSplits = randomSplits)
        forest.append(decisionTree)
    return forest

def random_forest_predict(dataFrame, randomForest):
    predictions = {}
    for i in range(len(randomForest)):
        column = "decision tree " + str(i)
        predictions[column] = decision_tree_predict(dataFrame, randomForest[i])
    predictions = pandas.DataFrame(predictions)
    return predictions.mode(axis = 1)[0]

def calculate_accuracy(prediction, category):
    correct = prediction == category
    return correct.mean()


start_time = datetime.datetime.now()

# https://www.tutorialspoint.com/python_data_access/python_mysql_database_connection.htm
# https://naysan.ca/2020/05/31/postgresql-to-pandas/
conn = psycopg2.connect(
   database="mqp", user='postgres', password='password', host='127.0.0.1', port= '5432'
)
cursor = conn.cursor()

#change the name "transfusion" to "adult" or "online" to test other datasets
cursor.execute("select * from transfusion")
tuples = cursor.fetchall()
cursor.close()
column_names = ['X1','X2','X3','X4','Y']

dataFrame = pd.DataFrame(tuples, columns=column_names)
conn.close()


dataFrameTrain, dataFrameTest = split_set(dataFrame, 0.2)

randomForest = random_forest(dataFrameTrain, bootstrap_size = 500, randomAttributes = 10, randomSplits = 3, forestSize = 10,
                                      treeMaxDepth = 5)
results = random_forest_predict(dataFrameTest, randomForest)
accuracy = calculateAccuracy(results, dataFrameTest.iloc[:, -1]) * 100

print("    accTest = {0:.2f}%, ".format(accuracy), end = "")


end_time = datetime.datetime.now()

time_diff = (end_time - start_time)
execution_time = time_diff.total_seconds() * 1000
print("Time: ", execution_time, "ms")
